@@ -1,26 +1,30 @@
-"""$Id: gdb.py.in,v 1.49 2006/07/29 08:09:32 rockyb Exp $
+"""$Id: gdb.py.in,v 1.90 2006/10/13 22:45:10 rockyb Exp $
 Handles gdb-like command processing.
 (See also pydb.doc for documentation.)"""
 
 _debugger_name = 'pydb'
 
 # The name of the debugger we are currently going by.
-_version = '1.17'
+_version = '1.19'
 
 import atexit, dis, inspect, linecache, os, pprint
-import bdb, re, subcmd, sys, traceback, types
+import bdb, re, subcmd, sys, types
 
 import pydbcmd, pydbbdb
+import signal
+import sighandler
 
-from display import Display, DisplayNode
-from fns     import *
-from info    import *
-from set     import *
-from show    import *
+## from connection import ConnectionServerFactory, ConnectionClientFactory, ConnectionFailed
+from display    import Display, DisplayNode
+from fns        import *
+from info       import *
+from set        import *
+from show       import *
 
 from pydbcmd import Cmd
 from pydbbdb import Bdb
 
+
 class Restart(Exception):
     """Causes a debugger to be restarted for the debugged Python program."""
     pass
@@ -37,14 +41,23 @@ class Gdb(Bdb, Cmd, SubcmdInfo, SubcmdSet, SubcmdShow):
         self._re_linetrace_delay  = re.compile(r'\s*linetrace\s+delay')
         self._wait_for_mainpyfile = False
 
-        self.__init_info()
-        self.__init_set()
-        self.__init_show()
+        # set up signal handling
+        self.sigmgr          = sighandler.SignalManager(self)
+        self._reset_handler  = None
+
+        self.__init_info()   # Initialize 'info' subcommands
+        self.__init_set()    # Initialize 'set' subcommands
+        self.__init_show()   # Initialize 'show' subcommands
 
         self.basename        = False
+        self.connection      = None   # For out-of-process communication
         self.currentbp       = None   # for "info program"
+        self.cur_frame       = None   # current stack frame
+        self.dbg_pydb        = False  # Debug the debugger?
+        self.debug_signal    = None   # The signal used by 'attach'
         self.display         = Display()
         self.gdb_dialect     = True   # Controls how stack is shown
+        self.lastcmd         = ''     # last debugger command run
         self.linetrace       = False
         self.linetrace_delay = 0
         self.listsize        = 10
@@ -54,17 +67,38 @@ class Gdb(Bdb, Cmd, SubcmdInfo, SubcmdSet, SubcmdShow):
 
         self.main_dirname    = os.curdir
         self.mainpyfile      = ''
-        self.maxargstrsize   = 30     # max length to show of parameter string
+        self.maxargstrsize   = 100     # max length to show of parameter string
         self.moduletodebug   = ""
-        self.noninteractive  = False
+        self.noninteractive  = False   # Controls whether to prompt on
+                                       # potentially dangerous commands.
         self.originalpath    = sys.path[:]
+        self.orig_stdin      = self.stdin
+        self.orig_stdout     = self.stdout
         self.running         = False
         self.search_path     = '$cdir:$cwd'  # source name search path
-        self.stop_reason     = None   # Why are we in the debugger?
-        self._sys_argv       = []     # exec sys.argv, e.g. may include pydb
+        self.sigcheck        = True
+        self.stop_reason     = None    # Why are we in the debugger?
+        self._sys_argv       = []      # exec sys.argv, e.g. may include pydb
         self.set_history_length   = None
+        self.stepping        = False   # used in thread debugging
 
 
+        self.target          = 'local' # local connections by default
+        self.target_addr     = ''      # target address used by 'attach'
+
+        # self.break_anywhere, self.set_continue and
+        # self.trace_dispatch are changed depending on the value of
+        # 'set sigcheck'.  We may "decorate", in the design pattern
+        # sense, the routines from bdb, so we need to save them.
+        self.break_anywhere_old  = self.break_anywhere
+        self.set_continue_old    = self.set_continue
+        self.trace_dispatch_old  = self.trace_dispatch
+
+        ## FIXME: for now we'll turn on signal trace dispatching
+        self.break_anywhere  = self.break_anywhere_gdb
+        self.set_continue    = self.set_continue_gdb
+        self.trace_dispatch  = self.trace_dispatch_gdb
+
         # associates a command list to breakpoint numbers
         self.commands = {}
 
@@ -113,6 +147,24 @@ class Gdb(Bdb, Cmd, SubcmdInfo, SubcmdSet, SubcmdShow):
         except ImportError:
             self.histfile = None
 
+    def break_anywhere_gdb(self, frame):
+        """Decorate bdb break_anyhwere to consider the sigcheck and
+        linetrace flags."""
+        if self.sigcheck or self.linetrace: return True
+        return self.break_anywhere(frame)
+
+    ### FIXME: trace_dispatch could be a performance bottleneck. If
+    ### so, consider reassigning the reoutine when don't have any
+    ### signals registered.
+            
+    def trace_dispatch_gdb(self, frame, event, arg):
+        """Check to see if the signal handler's we are interested have
+        changed. If so we'll intercept them. """
+        if not hasattr(self, 'thread_name') \
+           or self.thread_name == 'MainThread':
+            self.sigmgr.check_and_adjust_sighandlers()
+        return self.trace_dispatch_old(frame, event, arg)
+
     def __adjust_frame(self, pos, absolute_pos):
         """Adjust stack frame by pos positions. If absolute_pos then
         pos is an absolute number. Otherwise it is a relative number.
@@ -156,19 +208,18 @@ class Gdb(Bdb, Cmd, SubcmdInfo, SubcmdSet, SubcmdShow):
            If no completion we return the empty list.
            """
         matches = []
-        try:
-            cmdlist = getattr(self, "_%s_cmds" % subcmd)
-            for word in cmdlist:
-                if word.startswith(text):
-                    matches.append("%s %s" % (subcmd, word))
-        except AttributeError:
-            pass
+        subcmd_name = "%scmds" % subcmd[0]
+        if hasattr(self, subcmd_name):
+            subcmd_obj = getattr(self, subcmd_name)
+            if hasattr(subcmd_obj, "cmdlist"):
+                for word in subcmd_obj.cmdlist:
+                    if word.startswith(text):
+                        matches.append("%s %s" % (subcmd[0], word))
         return matches
 
-    def __do_complete(self, arg):
-        """Print a list of command names that can start with the
-        supplied command prefix. If there is no completion nothing
-        is printed."""
+    def all_completions(self, arg):
+        """Return a list of command names that can start with the
+        supplied command prefix."""
         if not arg:
             cmd_prefix=''
             args=[]
@@ -176,128 +227,92 @@ class Gdb(Bdb, Cmd, SubcmdInfo, SubcmdSet, SubcmdShow):
             args = arg.split()
             cmd_prefix=args[0]
         if len(args) > 1:
-            if args[0] in pydbcmd.Cmd.complete(self, cmd_prefix, 0):
-                text = self.__complete_subcommand(args, args[1])
-                for complete in text:  self.msg(complete)
+            complete_cmds = pydbcmd.Cmd.complete(self, cmd_prefix, 0)
+            if complete_cmds is not None and args[0] in complete_cmds:
+                return self.__complete_subcommand(args, args[1])
             return
 
         seen = {};
         i=0
+        completions = []
         text=pydbcmd.Cmd.complete(self, cmd_prefix, i)
         while True:
             if text is None: break
             if text not in seen.keys():
-                self.msg(text)
+                completions.append(text)
                 seen[text]=True
             i += 1
             text=pydbcmd.Cmd.complete(self, cmd_prefix, i)
+        return completions
 
-    def __get_brkpt_lineno(self, arg):
-        """__get_brkpt_lineno(self,arg)->(filename, file, lineno)
-
-        See if arg is a line number or a function name.  Return what
-        we've found. None can be returned as a value in the triple."""
-        funcname, filename = (None, None)
-        try:
-            # First try as an integer
-            lineno = int(arg)
-            filename = self.curframe.f_code.co_filename
-        except ValueError:
-            try:
-                func = eval(arg, self.curframe.f_globals,
-                            self.curframe.f_locals)
-            except:
-                func = arg
-            try:
-                if hasattr(func, 'im_func'):
-                    func = func.im_func
-                code = func.func_code
-                #use co_name to identify the bkpt (function names
-                #could be aliased, but co_name is invariant)
-                funcname = code.co_name
-                lineno = code.co_firstlineno
-                filename = code.co_filename
-            except:
-                # last thing to try
-                (ok, filename, ln) = self.lineinfo(arg)
-                if not ok:
-                    self.errmsg(('The specified object %s is not'
-                                 +' a function, or not found'
-                                 +' along sys.path or no line given.')
-                                % str(repr(arg)))
-                    return (None, None, None)
-                funcname = ok # ok contains a function name
-                lineno = int(ln)
-        return (funcname, filename, lineno)
+    def __do_complete(self, arg):
+        """Print a list of command names that can start with the
+        supplied command prefix. If there is no completion nothing
+        is printed."""
+        completions = self.all_completions(arg)
+        if completions is None: return
+        for complete in completions:  self.msg(complete)
 
     def __init_info(self):
-        """Initialize info subcommands"""
+        """Initialize info subcommands. Note: instance variable name
+        has to be infocmds ('info' + 'cmds') for subcommand completion
+        to work."""
         self.infocmds=subcmd.Subcmd("info", self.help_info.__doc__)
-        self.infocmds.add('args',         self.info_args)
-        self.infocmds.add('breakpoints',  self.info_breakpoints)
-        self.infocmds.add('display',      self.info_display)
-        self.infocmds.add('globals',      self.info_globals, 1, False)
-        self.infocmds.add('line',         self.info_line)
-        self.infocmds.add('locals',       self.info_locals,  1, False)
-        self.infocmds.add('program',      self.info_program)
-        self.infocmds.add('source',       self.info_source)
+        self.infocmds.add('args',           self.info_args)
+        self.infocmds.add('breakpoints',    self.info_breakpoints)
+        self.infocmds.add('display',        self.info_display)
+        self.infocmds.add('handle',         self.sigmgr.info_signal, 1, False)
+        self.infocmds.add('globals',        self.info_globals, 1, False)
+        self.infocmds.add('line',           self.info_line)
+        self.infocmds.add('locals',         self.info_locals,  1, False)
+        self.infocmds.add('program',        self.info_program)
+        self.infocmds.add('signal',         self.sigmgr.info_signal, 2, False)
+        self.infocmds.add('source',         self.info_source, 2)
+        ## self.infocmds.add('target',         self.info_target)
 
     def __init_set(self):
-        """Initialize set subcommands"""
-        self.setcmds=subcmd.Subcmd("set", self.help_set.__doc__)
-        self.setcmds.add('args',         self.set_args)
-        self.setcmds.add('basename',     self.set_basename)
-        self.setcmds.add('cmdtrace',     self.set_cmdtrace)
-        self.setcmds.add('history',      self.set_history)
-        self.setcmds.add('interactive',  self.set_interactive)
-        self.setcmds.add('linetrace',    self.set_linetrace, 3)
-        self.setcmds.add('listsize',     self.set_listsize,  3)
-        self.setcmds.add('logging',      self.set_logging,   2)
-        self.setcmds.add('prompt',       self.set_prompt)
+        """Initialize set subcommands. Note: instance variable name
+        has to be setcmds ('set' + 'cmds') for subcommand completion
+        to work."""
+
+        self.setcmds=subcmd.Subcmd("set",   self.help_set.__doc__)
+        self.setcmds.add('args',            self.set_args)
+        self.setcmds.add('basename',        self.set_basename)
+        self.setcmds.add('cmdtrace',        self.set_cmdtrace)
+        self.setcmds.add('debug-pydb',      self.set_dbg_pydb)
+        ## self.setcmds.add('debug-signal',    self.set_debug_signal)
+        self.setcmds.add('history',         self.set_history)
+        self.setcmds.add('interactive',     self.set_interactive)
+        self.setcmds.add('linetrace',       self.set_linetrace, 3)
+        self.setcmds.add('listsize',        self.set_listsize,  3)
+        self.setcmds.add('logging',         self.set_logging,   2)
+        self.setcmds.add('maxargsize',      self.set_maxargsize, 2)
+        self.setcmds.add('prompt',          self.set_prompt)
+        self.setcmds.add('sigcheck',        self.set_sigcheck)
+        ## self.setcmds.add('target-address',  self.set_target_address)
 
     def __init_show(self):
-        """Initialize show subcommands"""
+        """Initialize show subcommands. Note: instance variable name
+        has to be setcmds ('set' + 'cmds') for subcommand completion
+        to work."""
         self.showcmds=subcmd.Subcmd("show", self.help_show.__doc__)
-        self.showcmds.add('args',        self.show_args)
-        self.showcmds.add('basename',    self.show_basename)
-        self.showcmds.add('cmdtrace',    self.show_cmdtrace,    2)
-        self.showcmds.add('commands',    self.show_commands,    2, False)
-        self.showcmds.add('directories', self.show_directories, 1)
-        self.showcmds.add('history',     self.show_history)
-        self.showcmds.add('interactive', self.show_interactive)
-        self.showcmds.add('linetrace',   self.show_linetrace, 3)
-        self.showcmds.add('listsize',    self.show_listsize,  3)
-        self.showcmds.add('logging',     self.show_logging,   2)
-        self.showcmds.add('prompt',      self.show_prompt)
-        self.showcmds.add('version',     self.show_version)
-
-
-    def __parse_filepos(self, arg):
-        """__parse_filepos(self,arg)->(fn, filename, lineno)
-
-        Parse arg as [filename:]lineno | function
-        Make sure it works for C:\foo\bar.py:12
-        """
-        colon = arg.rfind(':')
-        if colon >= 0:
-            filename = arg[:colon].rstrip()
-            f = self.lookupmodule(filename)
-            if not f:
-                self.errmsg("%s not found from sys.path" %
-                            self._saferepr(filename))
-                return (None, None, None)
-            else:
-                filename = f
-                arg = arg[colon+1:].lstrip()
-            try:
-                lineno = int(arg)
-            except ValueError:
-                self.errmsg('Bad lineno: %s', str(arg))
-                return (None, filename, None)
-            return (None, filename, lineno)
-        else:
-            # no colon; can be lineno or function
-            return self.__get_brkpt_lineno(arg)
+        self.showcmds.add('args',           self.show_args)
+        self.showcmds.add('basename',       self.show_basename)
+        self.showcmds.add('debug-pydb',     self.show_dbg_pydb)
+        ## self.showcmds.add('debug-signal',   self.show_debug_signal)
+        self.showcmds.add('cmdtrace',       self.show_cmdtrace,    2)
+        self.showcmds.add('commands',       self.show_commands,    2, False)
+        self.showcmds.add('directories',    self.show_directories, 1)
+        self.showcmds.add('history',        self.show_history)
+        self.showcmds.add('interactive',    self.show_interactive)
+        self.showcmds.add('linetrace',      self.show_linetrace,   3)
+        self.showcmds.add('listsize',       self.show_listsize,    3)
+        self.showcmds.add('logging',        self.show_logging,     2)
+        self.showcmds.add('maxargsize',     self.show_maxargsize,  2)
+        self.showcmds.add('prompt',         self.show_prompt)
+        self.showcmds.add('sigcheck',       self.show_sigcheck)
+        ## self.showcmds.add('target-address', self.show_target_address)
+        self.showcmds.add('version',        self.show_version)
 
     # To be overridden in derived debuggers
     def defaultFile(self):
@@ -376,11 +391,76 @@ class Gdb(Bdb, Cmd, SubcmdInfo, SubcmdSet, SubcmdShow):
             # Tell cmdloop to break out of its loop.
             return True
 
+    def parse_list_cmd(self, arg):
+        """Parses arguments for the "list" command and returns the tuple
+        filename, start, last
+        or sets these to None if there was some problem."""
+
+        if not self.curframe:
+            self.msg("No Python program loaded.")
+            return (None, None, None)
+
+        self.lastcmd = 'list'
+        last = None
+        filename = self.curframe.f_code.co_filename
+        if arg:
+            if arg == '-':
+                first = max(1, self.lineno - 2*self.listsize - 1)
+            else:
+                args = arg.split()
+                (funcname, filename, first) = parse_filepos(self, args[0])
+                if first == None and funcname == None:
+                    # error should have been shown previously
+                    return (None, None, None)
+                if len(args) == 1:
+                    first = max(1, first - (self.listsize/2))
+                elif len(args) == 2:
+                    (funcname, last_filename, last) = \
+                               parse_filepos(self, args[1])
+                    if filename != last_filename \
+                       and filename is not None and last_filename is not None:
+                        self.errmsg('filename in the range must be the same')
+                        return (None, None, None)
+                    if last < first:
+                        try:
+                            int(args[1])
+                            # Assume it's a count
+                            last = first + last
+                        except ValueError:
+                            self.errmsg('2nd number (%s) is less than the '
+                                        + '1st number %d is not a pure integer'
+                                        % args[1], first)
+                            return (None, None, None)
+                elif len(args) > 2:
+                    self.errmsg('At most 2 arguments allowed. Saw: %d' %
+                                len(args))
+                    return (None, None, None)
+        elif self.lineno is None or not self.running:
+            first = max(1, inspect.getlineno(self.curframe) - self.listsize/2)
+        else:
+            first = self.lineno + 1
+        if last is None:
+            last = first + self.listsize - 1
+
+        return (filename, first, last)
+
+    def set_continue_gdb(self):
+        """Like bdb's set_continue but we don't have the run fast
+        option"""
+        self.stopframe = self.botframe
+        self.returnframe = None
+        self.quitting = 0
+
     def setup(self, frame, traceback=None):
         """Initialization done before entering the debugger-command
-        loop. We return 1 if we should NOT enter the debugger-command
+        loop. In particular we set up the call stack used for local
+        variable lookup and frame/up/down commands.
+
+        We return True if we should NOT enter the debugger-command
         loop."""
         self.forget()
+        if self.dbg_pydb:
+            frame = inspect.currentframe()
         if frame:
             self.stack, self.curindex = self.get_stack(frame, traceback)
             self.curframe = self.stack[self.curindex][0]
@@ -520,8 +600,8 @@ alias ps pi self
         else:
             self.aliases[args[0]] = ' '.join(args[1:])
 
-    def do_break(self, arg, temporary = 0):
-        """b(reak) {[file:]lineno | function} [, condition]
+    def do_break(self, arg, temporary = 0, thread_name=None):
+        """b(reak) [[file:]lineno | function] [, condition]
 With a line number argument, set a break there in the current
 file.  With a function name, set a break at first executable line
 of that function.  Without argument, list all breaks.  If a second
@@ -553,7 +633,7 @@ the .py suffix may be omitted."""
                 # parse stuff after comma: "condition"
                 cond = arg[comma+1:].lstrip()
                 arg = arg[:comma].rstrip()
-            (funcname, filename, lineno) = self.__parse_filepos(arg)
+            (funcname, filename, lineno) = parse_filepos(self, arg)
             if lineno is None: return
 
         # FIXME This default setting doesn't match that used in
@@ -576,8 +656,14 @@ the .py suffix may be omitted."""
             if err: self.errmsg(err)
             else:
                 bp = self.get_breaks(filename, line)[-1]
-                self.msg("Breakpoint %d set in file %s, line %d."
-                         % (bp.number, self.filename(bp.file), bp.line))
+                if thread_name is None:
+                    self.msg("Breakpoint %d set in file %s, line %d."
+                             % (bp.number, self.filename(bp.file), bp.line))
+                else:
+                    bp.thread_name = thread_name
+                    self.msg("Breakpoint %d set in file %s, line %d, thread %s."
+                             % (bp.number, self.filename(bp.file), bp.line,
+                                thread_name))
 
     do_b = do_break
 
@@ -624,7 +710,7 @@ the .py suffix may be omitted."""
                     self.errmsg("Invalid line number (%s)" % arg)
                     return
             else:
-                (funcname, filename, lineno) = self.__get_brkpt_lineno(arg)
+                (funcname, filename, lineno) = get_brkpt_lineno(self, arg)
 
             if lineno is None: return
 
@@ -717,9 +803,12 @@ the .py suffix may be omitted."""
     # That is the part that starts Continue execution...
     # So make sure this is a complete sentence.
     def do_continue(self, arg):
-        """c(ontinue)
+        """c(ontinue) [[file:]lineno | function]
 
-        Continue execution; only stop when a breakpoint is encountered."""
+Continue execution; only stop when a breakpoint is encountered. If a
+line position is given, continue until that line is reached. This is
+exactly the same thing as setting a temporary breakpoint at that
+position before running an (unconditional) continue."""
         if not self.is_running(): return
         if self.linetrace:
             # linetracing is like stepping, but we just don't stop. If
@@ -727,9 +816,16 @@ the .py suffix may be omitted."""
             # stopping if there were no breakpoints.
             self.step_ignore = -1
             self.set_step()
+            self.stepping = True
         else:
+            self.stepping = False
             self.set_continue()
 
+        if arg:
+            (funcname, filename, lineno) = parse_filepos(self, arg)
+            if lineno:
+                self.do_tbreak(arg)
+
         # Tell cmdloop to break out of its loop.
         return True
 
@@ -746,7 +842,7 @@ the .py suffix may be omitted."""
         sys.settrace(None)
         global_vars = self.curframe.f_globals
         local_vars = self.curframe.f_locals
-        p = Pdb()
+        p = Gdb()
         p.prompt  = "(%s) " % self.prompt.strip()
         self.msg("ENTERING RECURSIVE DEBUGGER")
 
@@ -755,10 +851,16 @@ the .py suffix may be omitted."""
         p.basename        = self.basename
         p.cmdtrace        = self.cmdtrace
         p.gdb_dialect     = self.gdb_dialect
+        p.infocmds        = self.infocmds
         p.linetrace       = self.linetrace
         p.linetrace_delay = self.linetrace_delay
         p.listsize        = self.listsize
         p.noninteractive  = self.noninteractive
+        for attr in ('thread_name', 'no_thread_do_break', 'nothread_do_tbreak',
+                     'nothread_trace_dispatch', 'nothread_quit',
+                     'desired_thread'):
+            if hasattr(self, attr):
+                setattr(p, attr, getattr(self, attr))
 
         # Some values that are different from the current environment
         # and different from the default initialization values.
@@ -767,9 +869,11 @@ the .py suffix may be omitted."""
 
         sys.call_tracing(p.run, (arg, global_vars, local_vars))
         self.msg("LEAVING RECURSIVE DEBUGGER")
+        # sys.settrace() seems to mess up self.print_location
+        # so print location first.
+        self.print_location()
         sys.settrace(self.trace_dispatch)
         self.lastcmd = p.lastcmd
-        self.print_location()
 
     def do_delete(self, arg):
         """delete [bpnumber [bpnumber...]]  - Delete some breakpoints.
@@ -849,7 +953,7 @@ the default."""
                 bp.disable()
 
     def do_disassemble(self, arg):
-        """disassemble [arg]
+        """disassemble [location]
         With no argument disassemble at the current frame location.
         With a numeric argument, disassemble at the frame location at that
         line number. With a class, method, function, code or string argument
@@ -964,45 +1068,31 @@ the default."""
 
     do_x = do_examine
 
-#    def do_file(self, fname):
-#
-#        """ Use FILE as program to be debugged.  It is compiled and
-#        becomes is the program executed when you use the `run'
-#        command.  If FILE cannot be found as specified, your execution
-#        directory path ($PATH) is searched for a command of that name.
-#        No arg means to have no executable file and no symbols."""
-#
-#        if fname == "":
-#            if self.mainpyfile == "":
-#                print 'No exec file now.\nNo symbol file now.'
-#                return
-#            else:
-#                # should confirm this per
-#                #Discard symbol table from `/work/.../ddd/cxxtest'? (y or n) n
-#                # XXX
-#                # del module
-#                sys.path = self.originalpath[:]
-#                self.mainpyfile == ""
-#        else:
-#            path, module = os.path.split(fname)
-#            module = module.split('.')[0]
-#            sys.path = self.originalpath[:]
-#            sys.path.insert(0, path)
-#            # XXX how to clean up name space ??
-#            if self.moduletodebug:
-#                try:
-#                    # del self.moduletodebug
-#                    self.moduletodebug = ''
-#                except:
-#                    pass
-#            line = "import " + module + '\n'
-#            code = compile(line, '<stdin>', 'single')
-#            global_vars = self.curframe.f_globals
-#            local_vars = self.curframe.f_locals
-#            exec code in global_vars, local_vars
-#            # Could print 'Reading symbols from fname...done.' for gdb conformance
-#            self.mainpyfile = fname
-#            self.moduletodebug = module
+    def do_file(self, fname):
+        """Use FILE as the Python program to be debugged.
+It is compiled and becomes is the program executed when you use the `run'
+command.  If no filename is given, this means to set things so there
+is no Python file."""
+        
+        if fname == "":
+           if self.mainpyfile == "":
+               self.msg('No exec file now.\nNo symbol file now.')
+               return
+           else:
+               # should confirm this per
+               if get_confirmation(self,
+                                   "Discard symbol table from %s'? (y or n)"
+                                   % self.mainpyfile):
+                   # XXX how to clean up name space ??
+                   self._program_sys_argv = None
+                   self.mainpyfile == ""
+        else:
+            # XXX how to clean up name space ??
+            self.mainpyfile=fname
+            self._program_sys_argv = [self.mainpyfile]
+            self._runscript(self.mainpyfile)
+            # Fake up curframe? 
+            return self.quitting
 
     def do_finish(self, arg):
         """finish
@@ -1010,9 +1100,11 @@ the default."""
         Continue execution until the current function returns."""
         if not self.is_running(): return
         self.set_return(self.curframe)
+        self.stepping = False
         if self.linetrace:
+            self.stepping    = True
             self.step_ignore = -1
-            self.stopframe = None
+            self.stopframe   = None
 
         # Tell cmdloop to break out of its loop.
         return True
@@ -1029,6 +1121,9 @@ the default."""
         So 'frame -1' moves when gdb dialect is in effect moves
         to the oldest frame, and 'frame 0' moves to the newest frame."""
 
+        if not self.stack:
+            self.msg("Program has no stack frame set.")
+            return
         try:
             arg = int(arg)
         except ValueError:
@@ -1043,6 +1138,22 @@ the default."""
         else:
             self.__adjust_frame(pos=arg, absolute_pos=True)
 
+    def do_handle(self, arg):
+        """Specify how to handle a signal.
+        
+Args are signals and actions to apply to those signals.
+recognized actions include "stop", "nostop", "print", "noprint",
+"pass", "nopass", "ignore", or "noignore".
+
+- Stop means reenter debugger if this signal happens (implies print and
+  nopass).
+- Print means print a message if this signal happens.
+- Pass means let program see this signal; otherwise program doesn't know.
+- Ignore is a synonym for nopass and noignore is a synonym for pass.
+- Pass and Stop may not be combined. (This is different from gdb)
+        """
+        self.sigmgr.action(arg)
+
     def do_ignore(self,arg):
         """ignore bpnumber count
 
@@ -1115,6 +1226,22 @@ the default."""
             except ValueError, e:
                 self.errmsg('%s failed: %s' % (cmdname, e))
 
+    def do_kill(self, arg):
+        """kill [unconditionally]
+
+Kill execution of program being debugged.
+
+Equivalent of kill -KILL <pid>. This is an unmaskable signal. When all
+else fails, e.g. in thread code, use this.
+
+If 'unconditionally' is given, no questions are asked. Otherwise, if
+we are in interactive mode, we'll prompt to make sure.
+"""
+        if len(arg) > 0 and 'unconditionally'.startswith(arg) or \
+               get_confirmation(self, 'Really do a hard kill (y or n)? ',
+                                True):
+            os.kill(os.getpid(), signal.SIGKILL)
+
     def do_list(self, arg):
         """l(ist) [- | first [last or count]]
 
@@ -1127,55 +1254,11 @@ the default."""
         first, it is a count.  First and last can be either a function
         name, a line number or file:line"""
 
-        if not self.curframe:
-            self.msg("No Python program loaded.")
-            return
-
-        self.lastcmd = 'list'
-        last = None
-        filename = self.curframe.f_code.co_filename
-        if arg:
-            if arg == '-':
-                first = max(1, self.lineno - 2*self.listsize - 1)
-            else:
-                args = arg.split()
-                (funcname, filename, first) = self.__parse_filepos(args[0])
-                if first == None and funcname == None:
-                    # error should have been shown previously
-                    return
-                if len(args) == 1:
-                    first = max(1, first - (self.listsize/2))
-                elif len(args) == 2:
-                    (funcname, last_filename, last) = \
-                               self.__parse_filepos(args[1])
-                    if filename != last_filename \
-                       and filename is not None and last_filename is not None:
-                        self.errmsg('filename in the range must be the same')
-                        return
-                    if last < first:
-                        try:
-                            int(args[1])
-                            # Assume it's a count
-                            last = first + last
-                        except ValueError:
-                            self.errmsg('2nd number (%s) is less than the '
-                                        + '1st number %d is not a pure integer'
-                                        % args[1], first)
-                            return
-                elif len(args) > 2:
-                    self.errmsg('At most 2 arguments allowed. Saw: %d' %
-                                len(args))
-                    return
-        elif self.lineno is None or not self.running:
-            first = max(1, inspect.getlineno(self.curframe) - self.listsize/2)
-        else:
-            first = self.lineno + 1
-        if last is None:
-            last = first + self.listsize - 1
+        filename, first, last = self.parse_list_cmd(arg)
+        if filename is None: return
         breaklist = self.get_file_breaks(filename)
 
-        # Okay we now have range information. Finally do the listing.
-        # (Just when you though we'd never get around to it).
+        # We now have range information. Do the listing.
         try:
             for lineno in range(first, last+1):
                 line = linecache.getline(filename, lineno)
@@ -1209,6 +1292,7 @@ With an integer argument, perform that many times."""
         except ValueError:
             return
 
+        self.stepping = True  # Used in thread debugging
         self.set_next(self.curframe)
         # Tell cmdloop to break out of its loop.
         return True
@@ -1237,7 +1321,14 @@ With an integer argument, perform that many times."""
 
     def do_quit(self, arg):
         """q(uit) or exit - Quit the debugger.  The program being
-        executed is aborted. For now, kill is a synonym for quit."""
+        executed is aborted."""
+        if self.target != 'local':
+            self._rebind_output(self.orig_stdout)
+            self._rebind_input(self.orig_stdin)
+            self._disconnect()
+            self.target = 'local'
+
+        sys.settrace(None)
         self._user_requested_quit = True
         self.running              = False
         self.set_quit()
@@ -1245,18 +1336,26 @@ With an integer argument, perform that many times."""
         # Tell cmdloop to break out of its loop.
         return True
 
-    do_kill = do_q = do_quit
+    do_q = do_quit
 
     def do_restart(self, arg):
         """restart - Restart debugger and program via an exec
         call. All state is lost, and new copy of the debugger is used."""
 
-        if self._sys_argv[0]:
-            self.msg("Re exec'ing\n\t%s" % self._sys_argv)
-            os.execvp(self._sys_argv[0], self._sys_argv)
+        # We don't proceed with the restart until the action has been
+        # ACK'd by any connected clients
+        if self.connection != None:
+            self.msg('restarting (connection)')
+            line = ""
+            while not 'ACK:restart_now' in line:
+                line = self.connection.readline()
+            self.do_rquit(None)
         else:
-            self.msg("No exectuable file specified.")
-
+            if self._sys_argv[0]:
+                self.msg("Re exec'ing\n\t%s" % self._sys_argv)
+                os.execvp(self._sys_argv[0], self._sys_argv)
+            else:
+                self.msg("No exectuable file specified.")
 
     def do_return(self, arg):
         """Make selected stack frame return to its caller. Control
@@ -1377,6 +1476,29 @@ assignment to the variable __return__."""
             args = arg.split()
             self.showcmds.do(self, args[0], args)
 
+    def do_signal(self, arg):
+        """signal [signum|signame]
+
+Send a signal to the debugged process.
+"""
+        if arg =='':
+            signum = 0
+        else:
+            try: 
+                signum = int(arg)
+                if sighandler.lookup_signame(signum) == None:
+                    self.msg("Signal number %d not a known signal number."
+                             % signum)
+                    return
+            except ValueError:
+                signum = sighandler.lookup_signum(arg)
+                if signum == None:
+                    self.msg("Signal name %s not a known signal name."
+                             % arg)
+                    return
+        os.kill(os.getpid(), signum)
+
+        
     def do_source(self, arg):
         """source FILE
         Read commands from a file named FILE.
@@ -1403,18 +1525,19 @@ With an integer argument, step that many times."""
         except ValueError:
             return
 
+        self.stepping = True  # Used in thread debugging
         self.set_step()
         # Tell cmdloop to break out of its loop.
         return True
 
     do_s = do_step
 
-    def do_tbreak(self, arg):
+    def do_tbreak(self, arg, thread_name=None):
         """tbreak  [ ([filename:]lineno | function) [, condition] ]
         Set a temporary breakpoint. Arguments are like the "break" command.
         Like "break" except the breakoint is only temporary,
         so it will be deleted when hit."""
-        self.do_break(arg, 1)
+        self.do_break(arg, 1, thread_name)
 
     def do_unalias(self, arg):
         """unalias name
@@ -1488,17 +1611,21 @@ Prints the type of the argument which can be a Python expression."""
             else:
                 self.errmsg("%s: %s" % (exc_type_name, self._saferepr(v)))
             return
-        if inspect.isfunction(value):
+        if inspect.ismethod(value):
+            self.msg('method %s%s' %
+                     (value.func_code.co_name,
+                       inspect.formatargspec(inspect.getargspec(value))))
+            if inspect.getdoc(value):
+                self.msg('%s:\n%s' %
+                         (value, inspect.getdoc(value)))
+            return
+        elif inspect.isfunction(value):
             self.msg('function %s%s' %
                      (value.func_code.co_name,
                        inspect.formatargspec(inspect.getargspec(value))))
             if inspect.getdoc(value):
                 self.msg('%s:\n%s' %
                          (value, inspect.getdoc(value)))
-                return
-        elif inspect.getdoc(value):
-            self.msg('%s:\n%s' %
-                     (value, inspect.getdoc(value)))
             return
         # None of the above...
         self.msg(type(value))
@@ -1538,7 +1665,7 @@ Handles the receipt of EOF as a command."""
     # Help methods (derived from pydb.doc or vice versa)
     #########################################################
 
-    for fn in ('EOF',     'L',         'alias',       'break',
+    for fn in ('EOF',     'alias',     'break',
                'cd',      'clear',     'condition',   'continue',
                'debug',   'disable',   'delete'   ,   'disassemble',
                'display', 'down',      'enable',      'examine',
@@ -1553,9 +1680,10 @@ Handles the receipt of EOF as a command."""
              % (fn, fn)
 
     # Is this the right way to do this?
-    help_h = help_help
-    help_R = help_run
-    help_bt = help_T = help_where
+    ## Remove duplicate short names
+    ## help_h = help_help
+    ## help_R = help_run
+    ## help_bt = help_T = help_where
 
     def help_commands(self, *arg):
         print """commands [bpnumber]
