@@ -1,4 +1,4 @@
-"""$Id: pydbbdb.py,v 1.18 2006/07/28 09:37:49 rockyb Exp $
+"""$Id: pydbbdb.py,v 1.24 2006/10/13 22:45:10 rockyb Exp $
 Routines here have to do with the subclassing of bdb.  Defines Python
 debugger Basic Debugger (Bdb) class.  This file could/should probably
 get merged into bdb.py
@@ -12,10 +12,12 @@ class Bdb(bdb.Bdb):
     def __init__(self):
         bdb.Bdb.__init__(self)
 
-        # A 0 value means stop on this occurance. A positive value means to
+        # A 0 value means stop on this occurrence. A positive value means to
         # skip that many more step/next's.
         self.step_ignore = 0
 
+        self.bdb_set_trace = bdb.Bdb.set_trace
+
         # Create a custom safe Repr instance and increase its maxstring.
         # The default of 30 truncates error messages too easily.
         self._repr = Repr()
@@ -26,6 +28,7 @@ class Bdb(bdb.Bdb):
         if self.linetrace:
             self.setup(frame)
             self.print_location(print_line=True)
+            self.display.displayAny(self.curframe)
             if self.linetrace_delay:
                 time.sleep(self.linetrace_delay)
 
@@ -107,6 +110,38 @@ class Bdb(bdb.Bdb):
             self.msg('\tbreakpoint already hit %d time%s' %
                      (bp.hits, ss), out)
 
+    def break_here(self, frame):
+        """This routine is almost copy of bdb.py's routine. Alas what pdb
+        calls clear gdb calls delete and gdb's clear command is different.
+        I tried saving/restoring method names, but that didn't catch
+        all of the places break_here was called.
+        """
+        filename = self.canonic(frame.f_code.co_filename)
+        if not filename in self.breaks:
+            return False
+        lineno = frame.f_lineno
+        if not lineno in self.breaks[filename]:
+            # The line itself has no breakpoint, but maybe the line is the
+            # first line of a function with breakpoint set by function name.
+            lineno = frame.f_code.co_firstlineno
+            if not lineno in self.breaks[filename]:
+                return False
+
+        # flag says ok to delete temp. bp
+        (bp, flag) = bdb.effective(filename, lineno, frame)
+        if bp:
+            ## This is new when we have thread debugging.
+            self.currentbp = bp.number
+            if hasattr(bp, 'thread_name') and hasattr(self, 'thread_name') \
+                   and bp.thread_name != self.thread_name:
+                    return False
+            if (flag and bp.temporary):
+                #### ARG. All for the below name change.
+                self.do_delete(str(bp.number))
+            return True
+        else:
+            return False
+
     def canonic(self, filename):
 
         """ Overrides bdb canonic. We need to ensure the file
@@ -216,10 +251,47 @@ class Bdb(bdb.Bdb):
         s += " %s at line %r" % (filename, lineno)
         return s
 
+    # The following two methods can be called by clients to use
+    # a debugger to debug a statement, given as a string.
+
+    def run(self, cmd, globals=None, locals=None):
+        """A copy of bdb's run but with a local variable added so we
+        can find it it a call stack and hide it when desired (which is
+        probably most of the time).
+        """
+        breadcrumb = self.run
+        if globals is None:
+            import __main__
+            globals = __main__.__dict__
+        if locals is None:
+            locals = globals
+        self.reset()
+        sys.settrace(self.trace_dispatch)
+        if not isinstance(cmd, types.CodeType):
+            cmd = cmd+'\n'
+        try:
+            self.running = True
+            try:
+                exec cmd in globals, locals
+            except bdb.BdbQuit:
+                pass
+        finally:
+            self.quitting = 1
+            self.running = False
+            sys.settrace(None)
+
     def reset(self):
         bdb.Bdb.reset(self)
         self.forget()
 
+    def set_trace(self, frame):
+        """Wrapper to accomodate different versions of Python"""
+        if sys.version_info[0] == 2 and sys.version_info[1] >= 4:
+            self.bdb_set_trace(self, self.curframe)
+        else:
+            # older versions
+            self.bdb_set_trace(self)
+
     def user_call(self, frame, argument_list):
         """This method is called when there is the remote possibility
         that we ever need to stop in this function."""
