@@ -1,8 +1,8 @@
-"""$Id: show.py,v 1.3 2006/07/25 14:15:43 rockyb Exp $
+"""$Id: show.py,v 1.10 2006/09/26 02:01:10 rockyb Exp $
 show subcommands, except those that need some sort of text substitution.
 (Those are in gdb.py.in.)
 """
-from fns import *
+import fns
 
 class SubcmdShow:
 
@@ -23,11 +23,24 @@ Follow this command with any number of args, to be passed to the program."""
 
     def show_basename(self, args):
         """Show if we are to show short of long filenames"""
-        self.msg("basename is %s." % show_onoff(self.basename))
+        self.msg("basename is %s." % fns.show_onoff(self.basename))
 
     def show_cmdtrace(self, args):
         "Show if we are to show debugger commands before running"
-        self.msg("cmdtrace is %s." % show_onoff(self.cmdtrace))
+        self.msg("cmdtrace is %s." % fns.show_onoff(self.cmdtrace))
+
+    def show_dbg_pydb(self, args):
+        """Show whether tracebacks include debugger routines"""
+        self.msg("Dbg_pydb is %s." % fns.show_onoff(self.dbg_pydb))
+
+    def show_debug_signal(self, arg):
+        """Show the signal currently used for triggering debugging
+        of an already running process.
+        """
+        if not self.debug_signal:
+            self.msg('debug-signal not set.')
+            return
+        self.msg('debug-signal is %s' % self.debug_signal)
 
     def show_directories(self, args):
         """Current search path for finding source files.
@@ -38,11 +51,16 @@ $cdir in the path means the compilation directory of the source file."""
     def show_interactive(self, args):
         """Show whether we are interactive"""
         self.msg("interactive is %s." %
-                 show_onoff(not self.noninteractive))
+                 fns.show_onoff(not self.noninteractive))
 
     def show_linetrace(self, args):
         "Show the line tracing status. Can also add 'delay'"
-        self.msg("line tracing is %s." % show_onoff(self.linetrace))
+        self.msg("line tracing is %s." % fns.show_onoff(self.linetrace))
+
+    def show_listsize(self, args):
+        """Show number of source lines the debugger will list by default."""
+        self.msg("Number of lines to show in listing is %s."
+                 %  self.listsize)
 
     def show_logging(self, args):
         "Show logging options"
@@ -53,10 +71,10 @@ $cdir in the path means the compilation directory of the source file."""
             elif args[1] == 'overwrite':
                 self.msg('Whether logging overwrites or appends to the'
                          + ' log file is %s.'
-                         % show_onoff(self.logging_overwrite))
+                         % fns.show_onoff(self.logging_overwrite))
             elif args[1] == 'redirect':
                 self.msg('The logging output mode is %s.' %
-                         show_onoff(self.logging_redirect))
+                         fns.show_onoff(self.logging_redirect))
             else:
                 self.undefined_cmd("show logging", args[1])
         else:
@@ -69,3 +87,26 @@ $cdir in the path means the compilation directory of the source file."""
                 self.msg("Output will be sent only to the log file.")
             else:
                 self.msg("Output will be logged and displayed.")
+
+    def show_maxargsize(self, args):
+        """Show number maximum number of characters in argument list."""
+        self.msg("Maximum number of characters in an argument list is %s"
+                 %  self.maxargstrsize)
+
+    def show_sigcheck(self, args):
+        """Show status of signal checking/adjusting.
+See also set sigcheck."""
+        self.msg("sigcheck is %s." % fns.show_onoff(self.sigcheck))
+
+    def show_target_address(self, arg):
+
+        """Show connection parameters used in remote debugging.
+
+This command doesn't make sense if you are not debugging a remote
+program. See also 'set target-address' and 'attach'."""
+
+        if self.target == 'local':
+            self.msg("Debugging is local. No target address.")
+        else:
+            self.msg('target-address is %s.' % self.target_addr.__repr__())
+
